from System import ArgumentException
from System.Drawing import (
    Font as WinFont,
    FontFamily,
    FontStyle,
    SystemFonts,
)
from System.Drawing.Text import PrivateFontCollection
from System.IO import FileNotFoundException
from System.Runtime.InteropServices import ExternalException

from toga.fonts import (
    _REGISTERED_FONT_CACHE,
    CURSIVE,
    FANTASY,
    MESSAGE,
    MONOSPACE,
    SANS_SERIF,
    SERIF,
    SYSTEM,
    SYSTEM_DEFAULT_FONT_SIZE,
)

_FONT_CACHE = {}

# Unlike SystemFonts.DefaultFont, MessageBoxFont respects the system theme
# (https://github.com/dotnet/winforms/issues/524).
DEFAULT_FONT = SystemFonts.MessageBoxFont


class Font:
    def __init__(self, interface):
        # this needs to be an instance variable, otherwise we might get Winforms
        # exceptions later
        self._pfc = None
        self.interface = interface
        try:
            font = _FONT_CACHE[self.interface]
        except KeyError:
            font = None
            font_key = self.interface._registered_font_key(
                self.interface.family,
                weight=self.interface.weight,
                style=self.interface.style,
                variant=self.interface.variant,
            )
            try:
                font_path = _REGISTERED_FONT_CACHE[font_key]
            except KeyError:
                try:
                    font_family = {
                        SYSTEM: DEFAULT_FONT.FontFamily,
                        MESSAGE: DEFAULT_FONT.FontFamily,
                        SERIF: FontFamily.GenericSerif,
                        SANS_SERIF: FontFamily.GenericSansSerif,
                        CURSIVE: FontFamily("Comic Sans MS"),
                        FANTASY: FontFamily("Impact"),
                        MONOSPACE: FontFamily.GenericMonospace,
                    }[self.interface.family]
                except KeyError:
                    try:
                        font_family = FontFamily(self.interface.family)
                    except ArgumentException:
                        print(
                            f"Unknown font '{self.interface}'; "
                            "using system font as a fallback"
                        )
                        font_family = DEFAULT_FONT.FontFamily

            else:
                try:
                    self._pfc = PrivateFontCollection()
                    self._pfc.AddFontFile(font_path)
                    font_family = self._pfc.Families[0]
                except FileNotFoundException:
                    raise ValueError(f"Font file {font_path} could not be found")
                except (IndexError, ExternalException):
                    raise ValueError(f"Unable to load font file {font_path}")

            # Convert font style to Winforms format
            font_style = FontStyle.Regular
            if self.interface.weight.lower() == "bold" and font_family.IsStyleAvailable(
                FontStyle.Bold
            ):
                font_style |= FontStyle.Bold
            if (
                # Winforms doesn't recognize Oblique; so we interpret as Italic
                self.interface.style.lower() in {"italic", "oblique"}
                and font_family.IsStyleAvailable(FontStyle.Italic)
            ):
                font_style |= FontStyle.Italic

            # Convert font size to Winforms format
            if self.interface.size == SYSTEM_DEFAULT_FONT_SIZE:
                font_size = DEFAULT_FONT.Size
            else:
                font_size = self.interface.size

            font = WinFont(font_family, font_size, font_style)
            _FONT_CACHE[self.interface] = font

        self.native = font

    def metric(self, name):
        """Return the given metric, measured in CSS pixels."""
        family = self.native.FontFamily
        style = self.native.Style
        em_height = self.native.SizeInPoints * 96 / 72
        design_unit = em_height / family.GetEmHeight(style)
        return design_unit * getattr(family, f"Get{name}")(style)



import System.Windows.Forms as WinForms

import toga
from toga.command import Group, Separator

from .libs.wrapper import WeakrefCallable


class StatusIcon:
    def __init__(self, interface):
        self.interface = interface
        self.native = None

    def set_icon(self, icon):
        if self.native:
            self.native.Icon = (
                icon._impl.native if icon else toga.App.app.icon._impl.native
            )

    def create(self):
        self.native = WinForms.NotifyIcon()
        self.native.Visible = True
        self.native.Text = self.interface.text
        self.set_icon(self.interface.icon)

    def remove(self):
        self.native.Visible = False
        self.native.Dispose()
        self.native = None


class SimpleStatusIcon(StatusIcon):
    def create(self):
        super().create()
        self.native.Click += WeakrefCallable(self.winforms_click)

    def winforms_click(self, sender, event):
        self.interface.on_press()


class MenuStatusIcon(StatusIcon):
    pass


class StatusIconSet:
    def __init__(self, interface):
        self.interface = interface
        self._menu_items = {}

    def _submenu(self, group, group_cache):
        try:
            return group_cache[group]
        except KeyError:
            if group is None:
                raise ValueError("Unknown top level item")
            else:
                parent_menu = self._submenu(group.parent, group_cache)

                submenu = WinForms.MenuItem(group.text)

                parent_menu.MenuItems.Add(submenu)

            group_cache[group] = submenu
        return submenu

    def create(self):
        # Menu status icons are the only icons that have extra construction needs.
        # Clear existing menus
        for item in self.interface._menu_status_icons:
            submenu = WinForms.ContextMenu()
            item._impl.native.ContextMenu = submenu

        # Determine the primary status icon.
        primary_group = self.interface._primary_menu_status_icon
        if primary_group is None:  # pragma: no cover
            # If there isn't at least one menu status icon, then there aren't any menus
            # to populate. This can't be replicated in the testbed.
            return

        # Add the menu status items to the cache
        group_cache = {
            item: item._impl.native.ContextMenu
            for item in self.interface._menu_status_icons
        }
        # Map the COMMANDS group to the primary status icon's menu.
        group_cache[Group.COMMANDS] = primary_group._impl.native.ContextMenu
        self._menu_items = {}

        for cmd in self.interface.commands:
            try:
                submenu = self._submenu(cmd.group, group_cache)
            except ValueError:
                raise ValueError(
                    f"Command {cmd.text!r} does not belong to "
                    "a current status icon group."
                )
            else:
                if isinstance(cmd, Separator):
                    menu_item = "-"
                else:
                    menu_item = cmd._impl.create_menu_item(WinForms.MenuItem)

                submenu.MenuItems.Add(menu_item)



import datetime
from decimal import ROUND_UP

import System.Windows.Forms as WinForms
from System import DateTime as WinDateTime

from ..libs.wrapper import WeakrefCallable
from .base import Widget


def py_time(native_time):
    return datetime.time(native_time.Hour, native_time.Minute, native_time.Second)


def native_time(py_time):
    # We don't need the date component, so we can use any date as long as we're
    # consistent.
    return WinDateTime(1970, 1, 1, py_time.hour, py_time.minute, py_time.second)


class TimeInput(Widget):

    def create(self):
        self.native = WinForms.DateTimePicker()
        self.native.ValueChanged += WeakrefCallable(self.winforms_value_changed)
        self.native.Format = WinForms.DateTimePickerFormat.Time
        self.native.ShowUpDown = True

    def get_value(self):
        return py_time(self.native.Value)

    def set_value(self, value):
        self.native.Value = native_time(value)

    def get_min_time(self):
        return py_time(self.native.MinDate)

    def set_min_time(self, value):
        self.native.MinDate = native_time(value)

    def get_max_time(self):
        return py_time(self.native.MaxDate)

    def set_max_time(self, value):
        self.native.MaxDate = native_time(value)

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )

    def winforms_value_changed(self, sender, event):
        self.interface.on_change()



import sys

from System.ComponentModel import InvalidEnumArgumentException

from toga import Command as StandardCommand, Group, Key
from toga_winforms.keys import toga_to_winforms_key, toga_to_winforms_shortcut
from toga_winforms.libs.wrapper import WeakrefCallable


class Command:
    def __init__(self, interface):
        self.interface = interface
        self.native = []

    @classmethod
    def standard(self, app, id):
        # ---- File menu -----------------------------------
        if id == StandardCommand.NEW:
            return {
                "text": "New",
                "shortcut": Key.MOD_1 + "n",
                "group": Group.FILE,
                "section": 0,
                "order": 0,
            }
        elif id == StandardCommand.OPEN:
            return {
                "text": "Open...",
                "shortcut": Key.MOD_1 + "o",
                "group": Group.FILE,
                "section": 0,
                "order": 10,
            }
        elif id == StandardCommand.SAVE:
            return {
                "text": "Save",
                "shortcut": Key.MOD_1 + "s",
                "group": Group.FILE,
                "section": 0,
                "order": 20,
            }
        elif id == StandardCommand.SAVE_AS:
            return {
                "text": "Save As...",
                "shortcut": Key.MOD_1 + "S",
                "group": Group.FILE,
                "section": 0,
                "order": 21,
            }
        elif id == StandardCommand.SAVE_ALL:
            return {
                "text": "Save All",
                "shortcut": Key.MOD_1 + Key.MOD_2 + "s",
                "group": Group.FILE,
                "section": 0,
                "order": 22,
            }
        elif id == StandardCommand.PREFERENCES:
            # Preferences should be towards the end of the File menu.
            return {
                "text": "Preferences",
                "group": Group.FILE,
                "section": sys.maxsize - 1,
            }
        elif id == StandardCommand.EXIT:
            # Quit should always be the last item, in a section on its own.
            return {
                "text": "Exit",
                "group": Group.FILE,
                "section": sys.maxsize,
            }
        # ---- Help menu -----------------------------------
        elif id == StandardCommand.VISIT_HOMEPAGE:
            return {
                "text": "Visit homepage",
                "enabled": app.home_page is not None,
                "group": Group.HELP,
            }
        elif id == StandardCommand.ABOUT:
            return {
                "text": f"About {app.formal_name}",
                "group": Group.HELP,
                "section": sys.maxsize,
            }

        raise ValueError(f"Unknown standard command {id!r}")

    def winforms_Click(self, sender, event):
        return self.interface.action()

    def set_enabled(self, value):
        if self.native:
            for widget in self.native:
                widget.Enabled = self.interface.enabled

    def create_menu_item(self, WinformsClass):
        item = WinformsClass(self.interface.text)

        item.Click += WeakrefCallable(self.winforms_Click)
        if self.interface.shortcut is not None:
            try:
                item.ShortcutKeys = toga_to_winforms_key(self.interface.shortcut)
                # The Winforms key enum is... daft. The "oem" key
                # values render as "Oem" or "Oemcomma", so we need to
                # *manually* set the display text for the key shortcut.
                item.ShortcutKeyDisplayString = toga_to_winforms_shortcut(
                    self.interface.shortcut
                )
            except (
                ValueError,
                InvalidEnumArgumentException,
            ) as e:  # pragma: no cover
                # Make this a non-fatal warning, because different backends may
                # accept different shortcuts.
                print(f"WARNING: invalid shortcut {self.interface.shortcut!r}: {e}")

        item.Enabled = self.interface.enabled

        self.native.append(item)

        return item



from System.Windows.Forms import TabControl

from .base import SimpleProbe


class OptionContainerProbe(SimpleProbe):
    native_class = TabControl
    max_tabs = None
    disabled_tab_selectable = True

    def select_tab(self, index):
        self.native.SelectedIndex = index

    async def wait_for_tab(self, message):
        await self.redraw(message)

    def tab_enabled(self, index):
        return self.native.TabPages[index].Enabled

    def assert_tab_icon(self, index, expected):
        # No tab icons, so if anything is returned, that's an error
        assert self.widget.content[index].icon is None



from Microsoft.Web.WebView2.WinForms import WebView2

from .base import SimpleProbe


class WebViewProbe(SimpleProbe):
    native_class = WebView2

    # https://github.com/MicrosoftEdge/WebView2Feedback/issues/530
    content_supports_url = False

    # https://github.com/MicrosoftEdge/WebView2Feedback/issues/983
    javascript_supports_exception = False

    supports_on_load = True

    def extract_cookie(self, cookie_jar, name):
        return next((c for c in cookie_jar if c.name == name), None)



import System.Windows.Forms as WinForms
from System.Drawing import ContentAlignment

from toga.constants import CENTER, JUSTIFY, LEFT, RIGHT


def TextAlignment(value):
    return {
        LEFT: ContentAlignment.TopLeft,
        RIGHT: ContentAlignment.TopRight,
        CENTER: ContentAlignment.TopCenter,
        JUSTIFY: ContentAlignment.TopLeft,
    }[value]


# Justify simply sets Left alignment. Is this the best option?
def HorizontalTextAlignment(value):
    return {
        LEFT: WinForms.HorizontalAlignment.Left,
        RIGHT: WinForms.HorizontalAlignment.Right,
        CENTER: WinForms.HorizontalAlignment.Center,
        JUSTIFY: WinForms.HorizontalAlignment.Left,
    }[value]



import asyncio
import re
import sys
import threading

import System.Windows.Forms as WinForms
from Microsoft.Win32 import SystemEvents
from System import Threading
from System.Media import SystemSounds
from System.Net import SecurityProtocolType, ServicePointManager
from System.Windows.Threading import Dispatcher

from .libs.proactor import WinformsProactorEventLoop
from .libs.wrapper import WeakrefCallable
from .screens import Screen as ScreenImpl


def winforms_thread_exception(sender, winforms_exc):  # pragma: no cover
    # The PythonException returned by Winforms doesn't give us
    # easy access to the underlying Python stacktrace; so we
    # reconstruct it from the string message.
    # The Python message is helpfully included in square brackets,
    # as the context for the first line in the .net stack trace.
    # So, look for the closing bracket and the start of the Python.net
    # stack trace. Then, reconstruct the line breaks internal to the
    # remaining string.
    print("Traceback (most recent call last):")
    py_exc = winforms_exc.get_Exception()
    full_stack_trace = py_exc.StackTrace
    regex = re.compile(
        r"^\[(?:'(.*?)', )*(?:'(.*?)')\]   (?:.*?) Python\.Runtime",
        re.DOTALL | re.UNICODE,
    )

    def print_stack_trace(stack_trace_line):  # pragma: no cover
        for level in stack_trace_line.split("', '"):
            for line in level.split("\\n"):
                if line:
                    print(line)

    stacktrace_relevant_lines = regex.findall(full_stack_trace)
    if len(stacktrace_relevant_lines) == 0:
        print_stack_trace(full_stack_trace)
    else:
        for lines in stacktrace_relevant_lines:
            for line in lines:
                print_stack_trace(line)

    print(py_exc.Message)


class App:
    # Winforms apps exit when the last window is closed
    CLOSE_ON_LAST_WINDOW = True
    # Winforms apps use default command line handling
    HANDLES_COMMAND_LINE = False

    def __init__(self, interface):
        self.interface = interface
        self.interface._impl = self

        # Track whether the app is exiting. This is used to stop the event loop,
        # and shortcut close handling on any open windows when the app exits.
        self._is_exiting = False

        # Winforms cursor visibility is a stack; If you call hide N times, you
        # need to call Show N times to make the cursor re-appear. Store a local
        # boolean to allow us to avoid building a deep stack.
        self._cursor_visible = True

        self.loop = WinformsProactorEventLoop()
        asyncio.set_event_loop(self.loop)

    def create(self):
        self.native = WinForms.Application
        self.app_context = WinForms.ApplicationContext()
        self.app_dispatcher = Dispatcher.CurrentDispatcher

        # We would prefer to detect DPI changes directly, using the DpiChanged,
        # DpiChangedBeforeParent or DpiChangedAfterParent events on the window. But none
        # of these events ever fire, possibly because we're missing some app metadata
        # (https://github.com/beeware/toga/pull/2155#issuecomment-2460374101). So
        # instead we need to listen to all events which could cause a DPI change:
        #   * DisplaySettingsChanged
        #   * Form.LocationChanged and Form.Resize, since a window's DPI is determined
        #     by which screen most of its area is on.
        SystemEvents.DisplaySettingsChanged += WeakrefCallable(
            self.winforms_DisplaySettingsChanged
        )

        # Ensure that TLS1.2 and TLS1.3 are enabled for HTTPS connections.
        # For some reason, some Windows installs have these protocols
        # turned off by default. SSL3, TLS1.0 and TLS1.1 are *not* enabled
        # as they are deprecated protocols and their use should *not* be
        # encouraged.
        try:
            ServicePointManager.SecurityProtocol |= SecurityProtocolType.Tls12
        except AttributeError:  # pragma: no cover
            print(
                "WARNING: Your Windows .NET install does not support TLS1.2. "
                "You may experience difficulties accessing some web server content."
            )
        try:
            ServicePointManager.SecurityProtocol |= SecurityProtocolType.Tls13
        except AttributeError:  # pragma: no cover
            print(
                "WARNING: Your Windows .NET install does not support TLS1.3. "
                "You may experience difficulties accessing some web server content."
            )

        # Populate the main window as soon as the event loop is running.
        self.loop.call_soon_threadsafe(self.interface._startup)

    ######################################################################
    # Native event handlers
    ######################################################################

    def winforms_DisplaySettingsChanged(self, sender, event):
        # This event is NOT called on the UI thread, so it's not safe for it to access
        # the UI directly.
        self.interface.loop.call_soon_threadsafe(self.update_dpi)

    def update_dpi(self):
        for window in self.interface.windows:
            window._impl.update_dpi()

    ######################################################################
    # Commands and menus
    ######################################################################

    def create_standard_commands(self):
        pass

    def create_menus(self):
        # Winforms menus are created on the Window.
        for window in self.interface.windows:
            # It's difficult to trigger this on a simple window, because we can't easily
            # modify the set of app-level commands that are registered, and a simple
            # window doesn't exist when the app starts up. Therefore, no-branch the else
            # case.
            if hasattr(window._impl, "create_menus"):  # pragma: no branch
                window._impl.create_menus()

    ######################################################################
    # App lifecycle
    ######################################################################

    def exit(self):  # pragma: no cover
        self._is_exiting = True
        self.native.Exit()

    def _run_app(self):  # pragma: no cover
        # Enable coverage tracing on this non-Python-created thread
        # (https://github.com/nedbat/coveragepy/issues/686).
        if threading._trace_hook:
            sys.settrace(threading._trace_hook)

        try:
            self.create()

            # This catches errors in handlers, and prints them
            # in a usable form.
            self.native.ThreadException += WeakrefCallable(winforms_thread_exception)

            self.loop.run_forever(self)
        except Exception as e:
            # In case of an unhandled error at the level of the app,
            # preserve the Python stacktrace
            self._exception = e
        else:
            # Ensure the event loop is fully closed.
            self.loop.close()
            self._exception = None

    def main_loop(self):
        thread = Threading.Thread(Threading.ThreadStart(self._run_app))
        thread.SetApartmentState(Threading.ApartmentState.STA)
        thread.Start()
        thread.Join()

        # If the thread has exited, the _exception attribute will exist.
        # If it's non-None, raise it, as it indicates the underlying
        # app thread had a problem; this is effectibely a re-raise over
        # a thread boundary.
        if self._exception:  # pragma: no cover
            raise self._exception

    def set_icon(self, icon):
        for window in self.interface.windows:
            window._impl.native.Icon = icon._impl.native

    def set_main_window(self, window):
        pass

    ######################################################################
    # App resources
    ######################################################################

    def get_primary_screen(self):
        return ScreenImpl(WinForms.Screen.PrimaryScreen)

    def get_screens(self):
        primary_screen = self.get_primary_screen()
        screen_list = [primary_screen] + [
            ScreenImpl(native=screen)
            for screen in WinForms.Screen.AllScreens
            if screen != primary_screen.native
        ]
        return screen_list

    ######################################################################
    # App state
    ######################################################################

    def get_dark_mode_state(self):
        self.interface.factory.not_implemented("dark mode state")
        return None

    ######################################################################
    # App capabilities
    ######################################################################

    def beep(self):
        SystemSounds.Beep.Play()

    def show_about_dialog(self):
        message_parts = []
        if self.interface.version is not None:
            message_parts.append(
                f"{self.interface.formal_name} v{self.interface.version}"
            )
        else:
            message_parts.append(self.interface.formal_name)

        if self.interface.author is not None:
            message_parts.append(f"Author: {self.interface.author}")
        if self.interface.description is not None:
            message_parts.append(f"\n{self.interface.description}")
        self.interface.main_window.info_dialog(
            f"About {self.interface.formal_name}", "\n".join(message_parts)
        )

    ######################################################################
    # Cursor control
    ######################################################################

    def hide_cursor(self):
        if self._cursor_visible:
            WinForms.Cursor.Hide()
        self._cursor_visible = False

    def show_cursor(self):
        if not self._cursor_visible:
            WinForms.Cursor.Show()
        self._cursor_visible = True

    ######################################################################
    # Window control
    ######################################################################

    def get_current_window(self):
        for window in self.interface.windows:
            if WinForms.Form.ActiveForm == window._impl.native:
                return window._impl
        return None

    def set_current_window(self, window):
        window._impl.native.Activate()



import System.Windows.Forms as WinForms

from .base import SimpleProbe


class ImageViewProbe(SimpleProbe):
    native_class = WinForms.PictureBox

    @property
    def preserve_aspect_ratio(self):
        return self.native.SizeMode == WinForms.PictureBoxSizeMode.Zoom

    def assert_image_size(self, width, height):
        # Winforms internally scales the image to the container,
        # so there's no image size check required.
        pass



import os
import platform
from pathlib import Path

import clr

#######################################################################
# WebView2
#######################################################################
WEBVIEW2_DIR = Path(__file__).parent / "WebView2"

# Derive the runtime's arch from the current machine's arch
arch_path = {
    "AMD64": "win-x64",
    "x86": "win-x86",
    "ARM64": "win-arm64",
}[platform.machine()]

# This specific filesystem layout is required for Windows Store Python.
# While python.org Python will respect $PATH when searching for DLLs to load,
# the Windows Store Python does not. Therefore, this filesystem layout is what
# the WebView2 loader will search by default to find the runtime.
# ref: https://learn.microsoft.com/en-us/microsoft-edge/webview2/concepts/distribution?tabs=dotnetcsharp#files-to-ship-with-the-app  # noqa: E501
webview_runtime_dir = WEBVIEW2_DIR / f"runtimes/{arch_path}/native"
os.environ["Path"] = f"{webview_runtime_dir}{os.pathsep}{os.environ['Path']}"

clr.AddReference(str(WEBVIEW2_DIR / "Microsoft.Web.WebView2.Core.dll"))
clr.AddReference(str(WEBVIEW2_DIR / "Microsoft.Web.WebView2.WinForms.dll"))

from Microsoft.Web.WebView2.Core import (  # noqa: F401, E402
    CoreWebView2Cookie,
    WebView2RuntimeNotFoundException,
)
from Microsoft.Web.WebView2.WinForms import (  # noqa: F401, E402
    CoreWebView2CreationProperties,
    WebView2,
)



from decimal import ROUND_UP

import System.Windows.Forms as WinForms

from toga.colors import TRANSPARENT
from toga.widgets.slider import IntSliderImpl

from ..libs.wrapper import WeakrefCallable
from .base import Widget

# Implementation notes
# ====================
#
# The native widget represents values as integers, so the IntSliderImpl base class is
# used to convert between integers and floats.


# "None" is a reserved word in Python, so we can't write WinForms.TickStyle.None.
NONE_TICK_STYLE = getattr(WinForms.TickStyle, "None")
BOTTOM_RIGHT_TICK_STYLE = WinForms.TickStyle.BottomRight


class Slider(Widget, IntSliderImpl):
    def create(self):
        IntSliderImpl.__init__(self)
        self.native = WinForms.TrackBar()
        self._default_background_color = TRANSPARENT
        self.native.AutoSize = False

        # Unlike Scroll, ValueChanged also fires when the value is changed
        # programmatically, such as via the testbed probe.
        self.native.ValueChanged += WeakrefCallable(self.winforms_value_changed)
        self.native.MouseDown += WeakrefCallable(self.winforms_mouse_down)
        self.native.MouseUp += WeakrefCallable(self.winforms_mouse_up)

    def winforms_value_changed(self, sender, event):
        self.on_change()

    def winforms_mouse_down(self, sender, event):
        self.interface.on_press()

    def winforms_mouse_up(self, sender, event):
        self.interface.on_release()

    def get_int_value(self):
        return self.native.Value

    def set_int_value(self, value):
        self.native.Value = value

    def get_int_max(self):
        return self.native.Maximum

    def set_int_max(self, value):
        self.native.Maximum = value

    def set_ticks_visible(self, visible):
        self.native.TickStyle = BOTTOM_RIGHT_TICK_STYLE if visible else NONE_TICK_STYLE

    def rehint(self):
        self.interface.intrinsic.height = self.scale_out(
            self.native.PreferredSize.Height, ROUND_UP
        )



from System.Drawing import Point
from System.Windows.Forms import Panel, ScrollEventArgs, ScrollEventType

from .base import SimpleProbe


class ScrollContainerProbe(SimpleProbe):
    native_class = Panel
    scrollbar_inset = 17

    def __init__(self, widget):
        super().__init__(widget)

        assert self.native.Controls.Count == 1
        self.native_content = self.native.Controls[0]
        assert isinstance(self.native_content, Panel)

    @property
    def has_content(self):
        return self.native_content.Controls.Count != 0

    @property
    def document_height(self):
        return round(self.native_content.Height / self.scale_factor)

    @property
    def document_width(self):
        return round(self.native_content.Width / self.scale_factor)

    async def scroll(self):
        if self.document_height > self.height:
            position = 100
            self.native.AutoScrollPosition = Point(0, position)
            self.native.OnScroll(
                ScrollEventArgs(ScrollEventType.ThumbPosition, position)
            )

    async def wait_for_scroll_completion(self):
        pass



from abc import ABC, abstractmethod
from datetime import date

from System import ArgumentOutOfRangeException
from System.Windows.Forms import DateTimePicker, DateTimePickerFormat

from .base import SimpleProbe


class DateTimeInputProbe(SimpleProbe, ABC):
    native_class = DateTimePicker
    fixed_height = 23
    supports_limits = True

    @abstractmethod
    def py_value(self, native_value):
        pass

    @property
    def value(self):
        return self.py_value(self.native.Value)

    @property
    def min_value(self):
        return self.py_value(self.native.MinDate)

    @property
    def max_value(self):
        return self.py_value(self.native.MaxDate)


class DateInputProbe(DateTimeInputProbe):
    def __init__(self, widget):
        super().__init__(widget)
        assert self.native.Format == DateTimePickerFormat.Long
        assert not self.native.ShowUpDown

    def py_value(self, native_value):
        return date(native_value.Year, native_value.Month, native_value.Day)

    async def change(self, delta):
        try:
            self.native.Value = self.native.Value.AddDays(delta)
        except ArgumentOutOfRangeException:
            pass
        await self.redraw(f"Change value by {delta} days")



import System.Windows.Forms
from System.Drawing import SystemColors

from .properties import toga_x_text_align
from .textinput import TextInputProbe


class MultilineTextInputProbe(TextInputProbe):
    native_class = System.Windows.Forms.RichTextBox
    fixed_height = None

    @property
    def value(self):
        return self.native.Text

    @property
    def value_hidden(self):
        return False

    @property
    def placeholder_visible(self):
        return self.native.ForeColor == SystemColors.GrayText

    @property
    def placeholder_hides_on_focus(self):
        return True

    def _char_pos(self, index):
        return self.native.GetPositionFromCharIndex(index)

    @property
    def document_height(self):
        text_len = len(self.native.Text)
        height = self._char_pos(text_len - 1).Y - self._char_pos(0).Y

        line_count = self.native.GetLineFromCharIndex(text_len - 1) + 1
        if line_count > 1:
            # Scale up to include the final line.
            assert height > 0
            height *= line_count / (line_count - 1)

        return round(height / self.scale_factor)

    @property
    def document_width(self):
        return self.width

    @property
    def vertical_scroll_position(self):
        return -round((self._char_pos(0).Y) / self.scale_factor)

    async def wait_for_scroll_completion(self):
        pass

    # According to the documentation: "SelectionAlignment returns
    # SelectionAlignment.Left when the text selection contains multiple paragraphs with
    # mixed alignment."
    @property
    def text_align(self):
        self.native.SelectAll()
        return toga_x_text_align(self.native.SelectionAlignment)



from __future__ import annotations

from typing import TYPE_CHECKING

import System.Windows.Forms as WinForms
from System.Drawing import Bitmap, Font as WinFont, Graphics, Point, Size as WinSize
from System.Drawing.Imaging import ImageFormat
from System.IO import MemoryStream

from toga import App
from toga.command import Separator
from toga.constants import WindowState
from toga.types import Position, Size

from .container import Container
from .fonts import DEFAULT_FONT
from .libs.wrapper import WeakrefCallable
from .screens import Screen as ScreenImpl
from .widgets.base import Scalable

if TYPE_CHECKING:  # pragma: no cover
    from toga.types import PositionT, SizeT


# It looks like something is caching the initial scale of the primary screen, and
# scaling all font sizes by it. Experiments show that this cache is at the level of the
# app, not the window.
initial_dpi_scale = ScreenImpl(WinForms.Screen.PrimaryScreen).dpi_scale


class Window(Container, Scalable):
    def __init__(self, interface, title, position, size):
        self.interface = interface

        self.create()

        self._FormClosing_handler = WeakrefCallable(self.winforms_FormClosing)
        self.native.FormClosing += self._FormClosing_handler
        super().__init__(self.native)
        self._dpi_scale = self.get_current_screen().dpi_scale

        self.native.MinimizeBox = self.interface.minimizable
        self.native.MaximizeBox = self.interface.resizable

        # Use a shadow variable since a window without any app menu and toolbar
        # in presentation mode would be indistinguishable from full screen mode.
        self._in_presentation_mode = False
        # Required to detect if the window has been un-minimized, and to prevent
        # double triggering of visibility events.
        self._previous_state = WindowState.NORMAL

        self.set_title(title)
        self.set_size(size)
        # Winforms does window cascading by default; use that behavior, rather than
        # Toga's re-implementation.
        if position:
            self.set_position(position)

        self.native.LocationChanged += WeakrefCallable(self.winforms_LocationChanged)
        self.native.Resize += WeakrefCallable(self.winforms_Resize)
        self.resize_content()  # Store initial size

        # Set window border style based on the window resizability setting at interface.
        self.native.FormBorderStyle = getattr(
            WinForms.FormBorderStyle,
            "Sizable" if self.interface.resizable else "FixedSingle",
        )

        self.native.Activated += WeakrefCallable(self.winforms_Activated)
        self.native.Deactivate += WeakrefCallable(self.winforms_Deactivate)
        self.native.VisibleChanged += WeakrefCallable(self.winforms_VisibleChanged)
        self.native.SizeChanged += WeakrefCallable(self.winforms_SizeChanged)

    def create(self):
        self.native = WinForms.Form()

    # We cache the scale to make sure that it only changes inside update_dpi.
    @property
    def dpi_scale(self):
        return self._dpi_scale

    def scale_font(self, native_font):
        return WinFont(
            native_font.FontFamily,
            native_font.Size * (self.dpi_scale / initial_dpi_scale),
            native_font.Style,
        )

    ######################################################################
    # Native event handlers
    ######################################################################

    def winforms_Resize(self, sender, event):
        if self.native.WindowState != WinForms.FormWindowState.Minimized:
            self.resize_content()

        # See DisplaySettingsChanged in app.py.
        if self.get_current_screen().dpi_scale != self._dpi_scale:
            self.update_dpi()

    def winforms_FormClosing(self, sender, event):
        # If the app is exiting, do nothing; we've already approved the exit
        # (and thus the window close). This branch can't be triggered in test
        # conditions, so it's marked no-branch.
        #
        # Otherwise, handle the close request by always cancelling the event,
        # and invoking `on_close()` handling. This will evaluate whether a close
        # is allowed, and if it is, programmatically invoke close on the window,
        # removing this handler first so that the close will complete.
        #
        # Winforms doesn't provide a way to disable/hide the close button, so if
        # the window is non-closable, don't trigger on_close handling - just
        # cancel the close event.
        if not self.interface.app._impl._is_exiting:  # pragma: no branch
            if self.interface.closable:
                self.interface.on_close()
            event.Cancel = True

    def winforms_LocationChanged(self, sender, event):
        # See DisplaySettingsChanged in app.py.
        if self.get_current_screen().dpi_scale != self._dpi_scale:
            self.update_dpi()

    def winforms_Activated(self, sender, event):
        self.interface.on_gain_focus()

    def winforms_Deactivate(self, sender, event):
        self.interface.on_lose_focus()

    def winforms_VisibleChanged(self, sender, event):
        if self.native.Visible:
            self.interface.on_show()
        else:
            self.interface.on_hide()

    def winforms_SizeChanged(self, sender, event):
        current_state = self.get_window_state()
        if self._previous_state != current_state:
            if self._previous_state == WindowState.MINIMIZED:
                self.interface.on_show()
            elif current_state == WindowState.MINIMIZED:
                self.interface.on_hide()
            self._previous_state = current_state

    ######################################################################
    # Window properties
    ######################################################################

    def get_title(self):
        return self.native.Text

    def set_title(self, title):
        self.native.Text = title

    ######################################################################
    # Window lifecycle
    ######################################################################

    def close(self):
        self.native.FormClosing -= self._FormClosing_handler
        self.native.Close()

    def set_app(self, app):
        icon_impl = app.interface.icon._impl
        self.native.Icon = icon_impl.native

    def show(self):
        if self.interface.content is not None:
            self.interface.content.refresh()
        self.update_dpi()
        self.native.Show()

    ######################################################################
    # Window content and resources
    ######################################################################

    # "Decor" includes the title bar and the (usually invisible) resize borders. It does
    # not include the menu bar and toolbar, which are included in the ClientSize (see
    # _top_bars_height).
    def _decor_width(self):
        return self.native.Size.Width - self.native.ClientSize.Width

    def _decor_height(self):
        return self.native.Size.Height - self.native.ClientSize.Height

    def _top_bars_height(self):
        return 0

    def refreshed(self):
        super().refreshed()
        layout = self.interface.content.layout
        self.native.MinimumSize = WinSize(
            self.scale_in(layout.min_width) + self._decor_width(),
            self.scale_in(layout.min_height)
            + self._top_bars_height()
            + self._decor_height(),
        )

    def resize_content(self):
        vertical_shift = self._top_bars_height()
        self.native_content.Location = Point(0, vertical_shift)
        super().resize_content(
            self.native.ClientSize.Width,
            self.native.ClientSize.Height - vertical_shift,
        )

    def update_dpi(self):
        self._dpi_scale = self.get_current_screen().dpi_scale

        # Update all the native fonts and determine the new preferred sizes.
        for widget in self.interface.widgets:
            widget._impl.scale_font()
            widget._impl.refresh()

        # Then do a single layout pass.
        if self.interface.content is not None:
            self.interface.content.refresh()

        self.resize_content()

    ######################################################################
    # Window size
    ######################################################################

    # Window.size is scaled according to the DPI of the current screen, to be consistent
    # with the scaling of its content.
    def get_size(self) -> Size:
        size = self.native.Size
        return Size(
            self.scale_out(size.Width - self._decor_width()),
            self.scale_out(size.Height - self._decor_height()),
        )

    def set_size(self, size: SizeT):
        self.native.Size = WinSize(
            self.scale_in(size[0]) + self._decor_width(),
            self.scale_in(size[1]) + self._decor_height(),
        )

    ######################################################################
    # Window position
    ######################################################################

    def get_current_screen(self):
        return ScreenImpl(WinForms.Screen.FromControl(self.native))

    # Window.position is scaled according to the DPI of the primary screen, because the
    # interface layer assumes that Screen.origin, Window.position and
    # Window.screen_position are all in the same coordinate system.
    #
    # TODO: remove that assumption, and make Window.position return coordinates relative
    # to the current screen's origin and DPI.
    def get_position(self) -> Position:
        location = self.native.Location
        primary_screen = App.app._impl.get_primary_screen()
        return Position(*map(primary_screen.scale_out, (location.X, location.Y)))

    def set_position(self, position: PositionT):
        primary_screen = App.app._impl.get_primary_screen()
        self.native.Location = Point(*map(primary_screen.scale_in, position))

    ######################################################################
    # Window visibility
    ######################################################################

    def get_visible(self):
        return self.native.Visible

    def hide(self):
        self.native.Hide()

    ######################################################################
    # Window state
    ######################################################################

    def get_window_state(self, in_progress_state=False):
        window_state = self.native.WindowState
        if window_state == WinForms.FormWindowState.Maximized:
            if self.native.FormBorderStyle == getattr(WinForms.FormBorderStyle, "None"):
                if self._in_presentation_mode:
                    return WindowState.PRESENTATION
                else:
                    return WindowState.FULLSCREEN
            else:
                return WindowState.MAXIMIZED
        elif window_state == WinForms.FormWindowState.Minimized:
            return WindowState.MINIMIZED
        else:  # window_state == WinForms.FormWindowState.Normal:
            return WindowState.NORMAL

    def set_window_state(self, state):
        # If the app is in presentation mode, but this window isn't, then
        # exit app presentation mode before setting the requested state.
        if any(
            window.state == WindowState.PRESENTATION and window != self.interface
            for window in self.interface.app.windows
        ):
            self.interface.app.exit_presentation_mode()

        current_state = self.get_window_state()
        if current_state == state:
            return

        elif current_state != WindowState.NORMAL:
            if current_state == WindowState.PRESENTATION:
                if self.native.MainMenuStrip:
                    self.native.MainMenuStrip.Visible = True
                if getattr(self, "toolbar_native", None):
                    self.toolbar_native.Visible = True

                self.interface.screen = self._before_presentation_mode_screen
                del self._before_presentation_mode_screen
                self._in_presentation_mode = False

            self.native.FormBorderStyle = getattr(
                WinForms.FormBorderStyle,
                "Sizable" if self.interface.resizable else "FixedSingle",
            )
            self.native.WindowState = WinForms.FormWindowState.Normal

            self.set_window_state(state)

        else:  # current_state == WindowState.NORMAL:
            if state == WindowState.MAXIMIZED:
                self.native.WindowState = WinForms.FormWindowState.Maximized

            elif state == WindowState.MINIMIZED:
                self.native.WindowState = WinForms.FormWindowState.Minimized

            elif state == WindowState.FULLSCREEN:
                self.native.FormBorderStyle = getattr(WinForms.FormBorderStyle, "None")
                self.native.WindowState = WinForms.FormWindowState.Maximized

            else:  # state == WindowState.PRESENTATION:
                self._before_presentation_mode_screen = self.interface.screen
                if self.native.MainMenuStrip:
                    self.native.MainMenuStrip.Visible = False
                if getattr(self, "toolbar_native", None):
                    self.toolbar_native.Visible = False
                self.native.FormBorderStyle = getattr(WinForms.FormBorderStyle, "None")
                self.native.WindowState = WinForms.FormWindowState.Maximized
                self._in_presentation_mode = True

    ######################################################################
    # Window capabilities
    ######################################################################

    def get_image_data(self):
        size = WinSize(self.native_content.Size.Width, self.native_content.Size.Height)
        bitmap = Bitmap(size.Width, size.Height)
        graphics = Graphics.FromImage(bitmap)

        graphics.CopyFromScreen(
            self.native_content.PointToScreen(Point.Empty),
            Point(0, 0),
            size,
        )

        stream = MemoryStream()
        bitmap.Save(stream, ImageFormat.Png)
        return bytes(stream.ToArray())


class MainWindow(Window):
    def create(self):
        super().create()
        self.toolbar_native = None

    def update_dpi(self):
        super().update_dpi()
        if self.native.MainMenuStrip:  # pragma: no branch
            self.native.MainMenuStrip.Font = self.scale_font(DEFAULT_FONT)
        if self.toolbar_native:
            self.toolbar_native.Font = self.scale_font(DEFAULT_FONT)
        self.resize_content()

    def _top_bars_height(self):
        vertical_shift = 0
        if self.toolbar_native and self.toolbar_native.Visible:
            vertical_shift += self.toolbar_native.Height
        if self.native.MainMenuStrip and self.native.MainMenuStrip.Visible:
            vertical_shift += self.native.MainMenuStrip.Height
        return vertical_shift

    def _submenu(self, group, group_cache):
        try:
            return group_cache[group]
        except KeyError:
            parent_menu = self._submenu(group.parent, group_cache)

            submenu = WinForms.ToolStripMenuItem(group.text)

            # Top level menus are added in a different way to submenus
            if group.parent is None:
                parent_menu.Items.Add(submenu)
            else:
                parent_menu.DropDownItems.Add(submenu)

            group_cache[group] = submenu
        return submenu

    def create_menus(self):
        menubar = self.native.MainMenuStrip
        if menubar:
            menubar.Items.Clear()
        else:
            # The menu bar doesn't need to be positioned, because its `Dock` property
            # defaults to `Top`.
            menubar = WinForms.MenuStrip()
            self.native.Controls.Add(menubar)
            self.native.MainMenuStrip = menubar
            self.native.MainMenuStrip.Font = self.scale_font(DEFAULT_FONT)
            menubar.SendToBack()  # In a dock, "back" means "top".

        group_cache = {None: menubar}

        submenu = None
        for cmd in self.interface.app.commands:
            submenu = self._submenu(cmd.group, group_cache)
            if isinstance(cmd, Separator):
                item = "-"
            else:
                item = cmd._impl.create_menu_item(WinForms.ToolStripMenuItem)

            submenu.DropDownItems.Add(item)

        self.resize_content()

    def create_toolbar(self):
        if self.interface.toolbar:
            if self.toolbar_native:
                self.toolbar_native.Items.Clear()
            else:
                # The toolbar doesn't need to be positioned, because its `Dock` property
                # defaults to `Top`.
                self.toolbar_native = WinForms.ToolStrip()
                self.native.Controls.Add(self.toolbar_native)
                self.toolbar_native.Font = self.scale_font(DEFAULT_FONT)
                self.toolbar_native.BringToFront()  # In a dock, "front" means "bottom".

            prev_group = None
            for cmd in self.interface.toolbar:
                if isinstance(cmd, Separator):
                    item = WinForms.ToolStripSeparator()
                    prev_group = None
                else:
                    # A change in group requires adding a toolbar separator
                    if prev_group is not None and prev_group != cmd.group:
                        self.toolbar_native.Items.Add(WinForms.ToolStripSeparator())
                        prev_group = None
                    else:
                        prev_group = cmd.group

                    item = WinForms.ToolStripButton(cmd.text)
                    item.AutoToolTip = False
                    if cmd.tooltip is not None:
                        item.ToolTipText = cmd.tooltip
                    if cmd.icon is not None:
                        item.Image = cmd.icon._impl.native.ToBitmap()
                    item.Enabled = cmd.enabled
                    item.Click += WeakrefCallable(cmd._impl.winforms_Click)
                    cmd._impl.native.append(item)
                self.toolbar_native.Items.Add(item)

        elif self.toolbar_native:
            self.native.Controls.Remove(self.toolbar_native)
            self.toolbar_native = None

        self.resize_content()
